@@ -240,9 +240,7 @@ defTag ide tag  =
     traceCTrav $ "Defining tag "++show ide++"...\n"
     otag <- transAttrCCT $ \ac -> addDefTagC ac ide tag
     case otag of
-      Nothing      -> do
-                        assertIfEnumThenFull tag
-                        return Nothing                  -- no collision
+      Nothing      -> return Nothing                  -- no collision
       Just prevTag -> case isRefinedOrUse prevTag tag of
                          Nothing                 -> return otag
                          Just (fullTag, foreIde) -> do
@@ -270,8 +268,11 @@ defTag ide tag  =
     isRefinedOrUse tag'@(StructUnionCT (CStruct _ (Just _  ) _  _ _))
                         (StructUnionCT (CStruct _ (Just ide') Nothing _ _)) =
       Just (tag', ide')
-    isRefinedOrUse tag'@(EnumCT        (CEnum (Just _  ) _  _ _))
-                        (EnumCT        (CEnum (Just ide') Nothing _ _))     =
+    isRefinedOrUse      (EnumCT        (CEnum (Just ide') Nothing _ _))
+                   tag'@(EnumCT        (CEnum (Just _  ) _  _ _)) =
+      Just (tag', ide')
+    isRefinedOrUse tag'@(EnumCT        (CEnum (Just ide') _ _ _))
+                        (EnumCT        (CEnum (Just _  ) _  _ _)) =
       Just (tag', ide')
     isRefinedOrUse _ _                                             = Nothing
 
@@ -293,9 +294,9 @@ findTagShadow ide  = readAttrCCT $ \ac -> lookupDefTagCShadow ac ide
 --   discarded, ie, all associations that existed before the transformation
 --   started are still in effect after the transformation
 --
-applyPrefixToNameSpaces        :: String -> CT s ()
-applyPrefixToNameSpaces prefix  =
-  transAttrCCT $ \ac -> (applyPrefix ac prefix, ())
+applyPrefixToNameSpaces        :: String -> String -> CT s ()
+applyPrefixToNameSpaces prefix repprefix  =
+  transAttrCCT $ \ac -> (applyPrefix ac prefix repprefix, ())
 
 -- definition attribute
 --
@@ -337,9 +338,9 @@ getDeclOf ide  =
       DontCareCD -> interr "CTrav.getDeclOf: Don't care!"
       TagCD _    -> interr "CTrav.getDeclOf: Illegal tag!"
       ObjCD obj  -> case obj of
-                      TypeCO    decl -> traceTypeCO >>
+                      TypeCO    decl -> traceTypeCO decl >>
                                         return decl
-                      ObjCO     decl -> traceObjCO >>
+                      ObjCO     decl -> traceObjCO decl >>
                                         return decl
                       EnumCO    _ _  -> illegalEnum
                       BuiltinCO      -> illegalBuiltin
@@ -353,10 +354,10 @@ getDeclOf ide  =
     traceEnter  = traceCTrav $
                     "Entering `getDeclOf' for `" ++ identToString ide
                     ++ "'...\n"
-    traceTypeCO = traceCTrav $
-                    "...found a type object.\n"
-    traceObjCO  = traceCTrav $
-                    "...found a vanilla object.\n"
+    traceTypeCO decl = traceCTrav $
+                    "...found a type object.\n" ++ show decl ++ "\n"
+    traceObjCO decl = traceCTrav $
+                    "...found a vanilla object.\n" ++ show decl ++ "\n"
 
 
 -- convenience functions
@@ -567,6 +568,12 @@ structFromDecl pos (CDecl specs _ _)  =
     CSUType su _ -> extractStruct pos (StructUnionCT su)
     _            -> structExpectedErr pos
 
+structFromDecl' :: Position -> CDecl -> CT s (Maybe CStructUnion)
+structFromDecl' pos (CDecl specs _ _)  =
+  case head [ts | CTypeSpec ts <- specs] of
+    CSUType su _ -> extractStruct' pos (StructUnionCT su)
+    _            -> structExpectedErr pos
+
 -- | extracts the arguments from a function declaration (must be a unique
 -- declarator) and constructs a declaration for the result of the function
 --
@@ -632,6 +639,7 @@ chaseDecl ide ind  =
 findAndChaseDecl                    :: Ident -> Bool -> Bool -> CT s CDecl
 findAndChaseDecl ide ind useShadows  =
   do
+    traceCTrav $ "findAndChaseDecl: " ++ show ide ++ "\n"
     (obj, ide') <- findTypeObj ide useShadows   -- is there an object def?
     ide  `refersToNewDef` ObjCD obj
     ide' `refersToNewDef` ObjCD obj             -- assoc needed for chasing
@@ -682,20 +690,23 @@ lookupEnum ide useShadows =
       Just (StructUnionCT _   ) -> enumExpectedErr ide  -- wrong tag definition
       Just (EnumCT        enum) -> return enum          -- enum tag definition
       Nothing                   -> do                   -- no tag definition
-        (CDecl specs _ _) <- findAndChaseDecl ide False useShadows
-        case head [ts | CTypeSpec ts <- specs] of
-          CEnumType enum _ -> return enum
-          _                -> enumExpectedErr ide
+        oobj <- if useShadows
+                then liftM (fmap fst) $ findObjShadow ide
+                else findObj ide
+        case oobj of
+          Just (EnumCO _ enum) -> return enum           -- anonymous enum
+          _                    -> do                    -- no value definition
+            (CDecl specs _ _) <- findAndChaseDecl ide False useShadows
+            case head [ts | CTypeSpec ts <- specs] of
+              CEnumType enum _ -> return enum
+              _                -> enumExpectedErr ide
 
 -- | for the given identifier, either find a struct/union in the tag name space
 -- or a type definition referring to a struct/union in the object name space;
 -- raises an error and exception if the identifier is not defined
 --
--- * if `ind = True', the identifier names a reference type to the searched
---   for struct/union
---
--- * typedef chasing is used only if there is no tag of the same name or an
---   indirection (ie, `ind = True') is explicitly required
+-- * the parameter `preferTag' determines whether tags or typedefs are
+--   searched first
 --
 -- * if the third argument is `True', use `findTagShadow'
 --
@@ -703,19 +714,37 @@ lookupEnum ide useShadows =
 --   definition
 --
 lookupStructUnion :: Ident -> Bool -> Bool -> CT s CStructUnion
-lookupStructUnion ide ind useShadows
-  | ind       = chase
-  | otherwise =
-    do
-      otag <- if useShadows
-              then liftM (fmap fst) $ findTagShadow ide
-              else findTag ide
-      maybe chase (extractStruct (posOf ide)) otag  -- `chase' if `Nothing'
-  where
-    chase =
-      do
-        decl <- findAndChaseDecl ide ind useShadows
+lookupStructUnion ide preferTag useShadows = do
+  traceCTrav $ "lookupStructUnion: ide=" ++ show ide ++ " preferTag=" ++
+    show preferTag ++ " useShadows=" ++ show useShadows ++ "\n"
+  otag <- if useShadows
+          then liftM (fmap fst) $ findTagShadow ide
+          else findTag ide
+  mobj <- if useShadows
+          then findObjShadow ide
+          else liftM (fmap (\obj -> (obj, ide))) $ findObj ide
+  let oobj = case mobj of
+        Just obj@(TypeCO _ , _) -> Just obj
+        Just obj@(BuiltinCO, _) -> Just obj
+        _                       -> Nothing
+  case preferTag of
+    True -> case otag of
+      Just tag -> extractStruct (posOf ide) tag
+      Nothing -> do
+        decl <- findAndChaseDecl ide True useShadows
         structFromDecl (posOf ide) decl
+    False -> case oobj of
+      Just _ -> do
+        decl <- findAndChaseDecl ide True useShadows
+        mres <- structFromDecl' (posOf ide) decl
+        case mres of
+          Just su -> return su
+          Nothing -> case otag of
+            Just tag -> extractStruct (posOf ide) tag
+            Nothing  -> unknownObjErr ide
+      Nothing -> case otag of
+        Just tag -> extractStruct (posOf ide) tag
+        Nothing  -> unknownObjErr ide
 
 -- | for the given identifier, check for the existance of both a type definition
 -- or a struct, union, or enum definition
@@ -785,19 +814,36 @@ extractAlias decl@(CDecl specs _ _) ind =
 --
 extractStruct                        :: Position -> CTag -> CT s CStructUnion
 extractStruct pos (EnumCT        _ )  = structExpectedErr pos
-extractStruct pos (StructUnionCT su)  =
+extractStruct pos (StructUnionCT su)  = do
+  traceCTrav $ "extractStruct: " ++ show su ++ "\n"
   case su of
     CStruct _ (Just ide') Nothing _ _ -> do            -- found forward definition
                                     def <- getDefOf ide'
+                                    traceCTrav $ "def=" ++ show def ++ "\n"
                                     case def of
                                       TagCD tag -> extractStruct pos tag
                                       bad_obj       -> err ide' bad_obj
     _                          -> return su
   where
-    err ide bad_obj = 
+    err ide bad_obj =
       do interr $ "CTrav.extractStruct: Illegal reference! Expected " ++ dumpIdent ide ++ 
                   " to link to TagCD but refers to "++ (show bad_obj) ++ "\n"
-                               
+
+extractStruct' :: Position -> CTag -> CT s (Maybe CStructUnion)
+extractStruct' pos (EnumCT        _ )  = structExpectedErr pos
+extractStruct' pos (StructUnionCT su)  = do
+  traceCTrav $ "extractStruct': " ++ show su ++ "\n"
+  case su of
+    CStruct _ (Just ide') Nothing _ _ -> do
+      def <- getDefOf ide'
+      traceCTrav $ "def=" ++ show def ++ "\n"
+      case def of
+        TagCD tag -> do
+          res <- extractStruct pos tag
+          return . Just $ res
+        _ -> return Nothing
+    _         -> return . Just $ su
+
 -- | yield the name declared by a declarator if any
 --
 declrName                          :: CDeclr -> Maybe Ident
@@ -814,13 +860,6 @@ assertFunDeclr pos (CDeclr _ (CFunDeclr _ _ _:retderiv) _ _ _) =
 assertFunDeclr pos _                                                 =
   funExpectedErr pos
 
--- | raise an error if the given tag defines an enumeration, but does not fully
--- define it
---
-assertIfEnumThenFull                          :: CTag -> CT s ()
-assertIfEnumThenFull (EnumCT (CEnum _ Nothing _ at))  = enumForwardErr (posOf at)
-assertIfEnumThenFull _                         = return ()
-
 -- | trace for this module
 --
 traceCTrav :: String -> CT s ()
@@ -873,9 +912,3 @@ structExpectedErr pos  =
   raiseErrorCTExc pos
     ["Expected a struct!",
      "Expected a structure or union; instead found an enum or basic type."]
-
-enumForwardErr     :: Position -> CT s a
-enumForwardErr pos  =
-  raiseErrorCTExc pos
-    ["Forward definition of enumeration!",
-     "ANSI C does not permit foreward definitions of enumerations!"]
