@@ -1,3 +1,4 @@
+{-# LANGUAGE StandaloneDeriving #-}
 --  C->Haskell Compiler: CHS file abstraction
 --
 --  Author : Manuel M T Chakravarty
@@ -51,16 +52,19 @@
 --            | `context' ctxt
 --            | `type' ident
 --            | `sizeof' ident
---            | `enum' idalias trans [`with' prefix] [deriving]
+--            | `alignof' ident
+--            | `enum' idalias trans [`with' prefix] [`add' prefix] [deriving]
+--            | `enum` `define` idalias [deriving]
 --            | `call' [`pure'] [`unsafe'] idalias
 --            | `fun' [`pure'] [`unsafe'] idalias parms
---            | `get' apath
---            | `set' apath
+--            | `get' [`struct'] apath
+--            | `set' [`struct'] apath
+--            | `offsetof` apath
 --            | `pointer' ['*'] idalias ptrkind ['nocode']
 --            | `class' [ident `=>'] ident ident
 --  ctxt     -> [`lib' `=' string] [prefix]
 --  idalias  -> ident [`as' (ident | `^')]
---  prefix   -> `prefix' `=' string
+--  prefix   -> `prefix' `=' string [`add' `prefix' `=' string]
 --  deriving -> `deriving' `(' ident_1 `,' ... `,' ident_n `)'
 --  parms    -> [verbhs `=>'] `{' parm_1 `,' ... `,' parm_n `}' `->' parm
 --  parm     -> [ident_or_quot_1 [`*' | `-']] verbhs [`&'] [ident_or_quot_2 [`*'] [`-']]
@@ -123,6 +127,15 @@ import C2HS.CHS.Lexer  (CHSToken(..), lexCHS, keywordToIdent)
 --
 data CHSModule = CHSModule [CHSFrag]
 
+deriving instance Show CHSModule
+deriving instance Show CHSFrag
+deriving instance Show CHSHook
+deriving instance Show CHSAccess
+deriving instance Show CHSParm
+deriving instance Show CHSTrans
+deriving instance Show CHSArg
+deriving instance Show CHSChangeCase
+
 -- | a CHS code fragament
 --
 -- * 'CHSVerb' fragments are present throughout the compilation and finally
@@ -141,8 +154,10 @@ data CHSModule = CHSModule [CHSFrag]
 data CHSFrag = CHSVerb String                   -- Haskell code
                        Position
              | CHSHook CHSHook                  -- binding hook
+                       Position
              | CHSCPP  String                   -- pre-processor directive
                        Position
+                       Bool
              | CHSLine Position                 -- line pragma
              | CHSC    String                   -- C code
                        Position
@@ -151,12 +166,12 @@ data CHSFrag = CHSVerb String                   -- Haskell code
                        (Maybe [CHSFrag])        -- else branch
 
 instance Pos CHSFrag where
-  posOf (CHSVerb _ pos ) = pos
-  posOf (CHSHook hook  ) = posOf hook
-  posOf (CHSCPP  _ pos ) = pos
-  posOf (CHSLine   pos ) = pos
-  posOf (CHSC    _ pos ) = pos
-  posOf (CHSCond alts _) = case alts of
+  posOf (CHSVerb _ pos  ) = pos
+  posOf (CHSHook _ pos  ) = pos
+  posOf (CHSCPP  _ pos _) = pos
+  posOf (CHSLine   pos  ) = pos
+  posOf (CHSC    _ pos  ) = pos
+  posOf (CHSCond alts _ ) = case alts of
                              (_, frag:_):_ -> posOf frag
                              _             -> nopos
 
@@ -168,6 +183,7 @@ data CHSHook = CHSImport  Bool                  -- qualified?
                           Position
              | CHSContext (Maybe String)        -- library name
                           (Maybe String)        -- prefix
+                          (Maybe String)        -- replacement prefix
                           Position
              | CHSType    Ident                 -- C type
                           Position
@@ -179,6 +195,7 @@ data CHSHook = CHSImport  Bool                  -- qualified?
                           (Maybe Ident)         -- Haskell name
                           CHSTrans              -- translation table
                           (Maybe String)        -- local prefix
+                          (Maybe String)        -- local replacement prefix
                           [Ident]               -- instance requests from user
                           Position
              | CHSEnumDefine Ident              -- Haskell name
@@ -201,6 +218,8 @@ data CHSHook = CHSImport  Bool                  -- qualified?
              | CHSField   CHSAccess             -- access type
                           CHSAPath              -- access path
                           Position
+             | CHSOffsetof CHSAPath             -- access path
+                           Position
              | CHSPointer Bool                  -- explicit '*' in hook
                           Ident                 -- C pointer name
                           (Maybe Ident)         -- Haskell name
@@ -216,15 +235,16 @@ data CHSHook = CHSImport  Bool                  -- qualified?
 
 instance Pos CHSHook where
   posOf (CHSImport  _ _ _         pos) = pos
-  posOf (CHSContext _ _           pos) = pos
+  posOf (CHSContext _ _ _         pos) = pos
   posOf (CHSType    _             pos) = pos
   posOf (CHSSizeof  _             pos) = pos
   posOf (CHSAlignof _             pos) = pos
-  posOf (CHSEnum    _ _ _ _ _     pos) = pos
+  posOf (CHSEnum    _ _ _ _ _ _   pos) = pos
   posOf (CHSEnumDefine _ _ _      pos) = pos
   posOf (CHSCall    _ _ _ _       pos) = pos
   posOf (CHSFun     _ _ _ _ _ _ _ pos) = pos
   posOf (CHSField   _ _           pos) = pos
+  posOf (CHSOffsetof _            pos) = pos
   posOf (CHSPointer _ _ _ _ _ _ _ pos) = pos
   posOf (CHSClass   _ _ _         pos) = pos
 
@@ -232,31 +252,32 @@ instance Pos CHSHook where
 -- same C object
 --
 instance Eq CHSHook where
-  (CHSImport qual1 ide1 _      _) == (CHSImport qual2 ide2 _      _) =
+  (CHSImport qual1 ide1 _           _) == (CHSImport qual2 ide2 _           _) =
     qual1 == qual2 && ide1 == ide2
-  (CHSContext olib1 opref1  _   ) == (CHSContext olib2 opref2  _   ) =
-    olib1 == olib2 && opref1 == opref2
-  (CHSType ide1                _) == (CHSType ide2                _) =
+  (CHSContext olib1 opref1 orpref1  _) == (CHSContext olib2 opref2 orpref2  _) =
+    olib1 == olib2 && opref1 == opref2 && orpref1 == orpref2
+  (CHSType ide1                     _) == (CHSType ide2                     _) =
     ide1 == ide2
-  (CHSSizeof ide1              _) == (CHSSizeof ide2              _) =
+  (CHSSizeof ide1                   _) == (CHSSizeof ide2                   _) =
     ide1 == ide2
-  (CHSAlignof ide1             _) == (CHSAlignof ide2             _) =
-    ide1 == ide2    
-  (CHSEnum ide1 oalias1 _ _ _  _) == (CHSEnum ide2 oalias2 _ _ _  _) =
-    oalias1 == oalias2 && ide1 == ide2
-  (CHSEnumDefine ide1 _ _      _) == (CHSEnumDefine ide2 _ _      _) =
+  (CHSAlignof ide1                  _) == (CHSAlignof ide2                  _) =
     ide1 == ide2
-  (CHSCall _ _ ide1 oalias1    _) == (CHSCall _ _ ide2 oalias2    _) =
+  (CHSEnum ide1 oalias1 _ _ _ _     _) == (CHSEnum ide2 oalias2 _ _ _ _     _) =
     oalias1 == oalias2 && ide1 == ide2
-  (CHSFun  _ _ ide1 oalias1 _ _ _ _)
-                                  == (CHSFun _ _ ide2 oalias2 _ _ _ _) =
+  (CHSEnumDefine ide1 _ _           _) == (CHSEnumDefine ide2 _ _           _) =
+    ide1 == ide2
+  (CHSCall _ _ ide1 oalias1         _) == (CHSCall _ _ ide2 oalias2         _) =
     oalias1 == oalias2 && ide1 == ide2
-  (CHSField acc1 path1         _) == (CHSField acc2 path2         _) =
+  (CHSFun  _ _ ide1 oalias1 _ _ _   _) == (CHSFun _ _ ide2 oalias2 _ _ _    _) =
+    oalias1 == oalias2 && ide1 == ide2
+  (CHSField acc1 path1              _) == (CHSField acc2 path2              _) =
     acc1 == acc2 && path1 == path2
+  (CHSOffsetof path1                _) == (CHSOffsetof path2                _) =
+    path1 == path2
   (CHSPointer _ ide1 oalias1 _ _ _ _ _)
-                                  == (CHSPointer _ ide2 oalias2 _ _ _ _ _) =
+                                      == (CHSPointer _ ide2 oalias2 _ _ _ _ _) =
     ide1 == ide2 && oalias1 == oalias2
-  (CHSClass _ ide1 _           _) == (CHSClass _ ide2 _           _) =
+  (CHSClass _ ide1 _                _) == (CHSClass _ ide2 _                _) =
     ide1 == ide2
   _                               == _                          = False
 
@@ -300,13 +321,15 @@ data CHSAccess = CHSSet                         -- set structure field
 
 -- | structure access path
 --
-data CHSAPath = CHSRoot  Ident                  -- root of access path
+data CHSAPath = CHSRoot Bool Ident
+                -- root of access path with flag indicating presence
+                -- of "struct" keyword
               | CHSDeref CHSAPath Position      -- dereferencing
               | CHSRef   CHSAPath Ident         -- member referencing
               deriving (Eq,Show)
 
 instance Pos CHSAPath where
-    posOf (CHSRoot ide)    = posOf ide
+    posOf (CHSRoot _ ide)  = posOf ide
     posOf (CHSDeref _ pos) = pos
     posOf (CHSRef _ ide)   = posOf ide
 
@@ -442,19 +465,19 @@ showCHSModule (CHSModule fragments) pureHaskell  =
       in
         (if emitNow then
            showString ("\n{-# LINE " ++ show (line `max` 0) ++ " " ++
-                       show fname ++ " #-}")
+                       show fname ++ " #-}\n")
          else id)
       . showString s
       . showFrags pureHs nextState frags
-    showFrags False  _     (CHSHook hook       : frags) =
+    showFrags False  _     (CHSHook hook _     : frags) =
         showString "{#"
       . showCHSHook hook
       . showString "#}"
       . showFrags False Wait frags
-    showFrags False  _     (CHSCPP  s    _     : frags) =
-        showChar '#'
+    showFrags False  _     (CHSCPP  s    _ nl  : frags) =
+      (if nl then showChar '\n' else id)
+      . showChar '#'
       . showString s
---      . showChar '\n'
       . showFrags False Emit frags
     showFrags pureHs _     (CHSLine _s         : frags) =
         showFrags pureHs Emit frags
@@ -473,12 +496,13 @@ showCHSHook (CHSImport isQual ide _ _) =
     showString "import "
   . (if isQual then showString "qualified " else id)
   . showCHSIdent ide
-showCHSHook (CHSContext olib oprefix _) =
+showCHSHook (CHSContext olib oprefix oreplprefix _) =
     showString "context "
   . (case olib of
        Nothing  -> showString ""
        Just lib -> showString "lib = " . showString lib . showString " ")
   . showPrefix oprefix False
+  . showReplacementPrefix oreplprefix
 showCHSHook (CHSType ide _) =
     showString "type "
   . showCHSIdent ide
@@ -488,11 +512,12 @@ showCHSHook (CHSSizeof ide _) =
 showCHSHook (CHSAlignof   ide _) =
    showString "alignof "
   . showCHSIdent ide
-showCHSHook (CHSEnum ide oalias trans oprefix derive _) =
+showCHSHook (CHSEnum ide oalias trans oprefix oreplprefix derive _) =
     showString "enum "
   . showIdAlias ide oalias
   . showCHSTrans trans
   . showPrefix oprefix True
+  . showReplacementPrefix oreplprefix
   . if null derive then id else showString $
       "deriving ("
       ++ concat (intersperse ", " (map identToString derive))
@@ -527,6 +552,9 @@ showCHSHook (CHSField acc path _) =
        CHSGet -> showString "get "
        CHSSet -> showString "set ")
   . showCHSAPath path
+showCHSHook (CHSOffsetof path _) =
+    showString "offsetof "
+  . showCHSAPath path
 showCHSHook (CHSPointer star ide oalias ptrType isNewtype oRefType emit _) =
     showString "pointer "
   . (if star then showString "*" else showString "")
@@ -560,6 +588,12 @@ showPrefix (Just prefix) withWith  =   maybeWith
   where
     maybeWith = if withWith then showString "with " else id
 
+showReplacementPrefix              :: Maybe String -> ShowS
+showReplacementPrefix Nothing       = showString ""
+showReplacementPrefix (Just prefix) = showString "add prefix = "
+                                      . showString prefix
+                                      . showString " "
+
 showIdAlias            :: Ident -> Maybe Ident -> ShowS
 showIdAlias ide oalias  =
     showCHSIdent ide
@@ -617,7 +651,10 @@ showCHSChangeCase CHSUpCase   = showString "upcaseFirstLetter"
 showCHSChangeCase CHSDownCase = showString "downcaseFirstLetter"
 
 showCHSAPath :: CHSAPath -> ShowS
-showCHSAPath (CHSRoot ide) =
+showCHSAPath (CHSRoot True ide) =
+    showString "struct "
+  . showCHSIdent ide
+showCHSAPath (CHSRoot False ide) =
   showCHSIdent ide
 showCHSAPath (CHSDeref path _) =
     showString "* "
@@ -777,25 +814,39 @@ parseFrags tokens  = do
     parseFrags0 (CHSTokCtrl    pos c:toks) = do
                                                frags <- parseFrags toks
                                                return $ CHSVerb [c] pos : frags
-    parseFrags0 (CHSTokCPP     pos s:toks) = do
+    parseFrags0 (CHSTokCPP     pos s nl:toks) = do
                                                frags <- parseFrags toks
-                                               return $ CHSCPP s pos : frags
+                                               return $ CHSCPP s pos nl : frags
     parseFrags0 (CHSTokLine    pos  :toks) = do
                                                frags <- parseFrags toks
                                                return $ CHSLine pos : frags
     parseFrags0 (CHSTokC       pos s:toks) = parseC       pos s      toks
-    parseFrags0 (CHSTokImport  pos  :toks) = parseImport  pos        toks
-    parseFrags0 (CHSTokContext pos  :toks) = parseContext pos        toks
-    parseFrags0 (CHSTokType    pos  :toks) = parseType    pos        toks
-    parseFrags0 (CHSTokSizeof  pos  :toks) = parseSizeof  pos        toks
-    parseFrags0 (CHSTokAlignof pos  :toks) = parseAlignof pos        toks
-    parseFrags0 (CHSTokEnum    pos  :toks) = parseEnum    pos        toks
-    parseFrags0 (CHSTokCall    pos  :toks) = parseCall    pos        toks
-    parseFrags0 (CHSTokFun     pos  :toks) = parseFun     pos        toks
-    parseFrags0 (CHSTokGet     pos  :toks) = parseField   pos CHSGet toks
-    parseFrags0 (CHSTokSet     pos  :toks) = parseField   pos CHSSet toks
-    parseFrags0 (CHSTokClass   pos  :toks) = parseClass   pos        toks
-    parseFrags0 (CHSTokPointer pos  :toks) = parsePointer pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokImport  pos  :toks) = parseImport  hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokContext pos  :toks) = parseContext hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokType    pos  :toks) = parseType    hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokSizeof  pos  :toks) = parseSizeof  hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokAlignof pos  :toks) = parseAlignof hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokEnum    pos  :toks) = parseEnum    hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokCall    pos  :toks) = parseCall    hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokFun     pos  :toks) = parseFun     hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokGet     pos  :toks) = parseField   hkpos pos CHSGet toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokSet     pos  :toks) = parseField   hkpos pos CHSSet toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokOffsetof pos :toks) = parseOffsetof hkpos pos       toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokClass   pos  :toks) = parseClass   hkpos pos        toks
+    parseFrags0 (CHSTokHook hkpos:
+                 CHSTokPointer pos  :toks) = parsePointer hkpos pos        toks
     parseFrags0 toks                       = syntaxError toks
     --
     -- skip to next Haskell or control token
@@ -819,8 +870,8 @@ parseC pos s toks =
                                                 return $ CHSC s'  pos' : frags
     collectCtrlAndC toks'                      = parseFrags toks'
 
-parseImport :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseImport pos toks = do
+parseImport :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseImport hkpos pos toks = do
   (qual, modid, toks') <-
     case toks of
       CHSTokIdent _ ide                :toks' ->
@@ -833,7 +884,7 @@ parseImport pos toks = do
   chi <- loadCHI . moduleNameToFileName . identToString $ modid
   toks'2 <- parseEndHook toks'
   frags <- parseFrags toks'2
-  return $ CHSHook (CHSImport qual modid chi pos) : frags
+  return $ CHSHook (CHSImport qual modid chi pos) hkpos : frags
 
 -- | Qualified module names do not get lexed as a single token so we need to
 -- reconstruct it from a sequence of identifer and dot tokens.
@@ -851,69 +902,72 @@ moduleNameToFileName = map dotToSlash
   where dotToSlash '.' = '/'
         dotToSlash c   = c
 
-parseContext          :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseContext pos toks  = do
-                           (olib    , toks'2 ) <- parseOptLib          toks
-                           (opref   , toks'3) <- parseOptPrefix False toks'2
-                           toks'4            <- parseEndHook         toks'3
-                           frags               <- parseFrags           toks'4
-                           let frag = CHSContext olib opref pos
-                           return $ CHSHook frag : frags
+parseContext          :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseContext hkpos pos toks  = do
+  (olib    , toks2) <- parseOptLib          toks
+  (opref   , toks3) <- parseOptPrefix False toks2
+  (oreppref, toks4) <- parseOptReplPrefix   toks3
+  toks5             <- parseEndHook         toks4
+  frags             <- parseFrags           toks5
+  let frag = CHSContext olib opref oreppref pos
+  return $ CHSHook frag hkpos : frags
 
-parseType :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseType pos (CHSTokIdent _ ide:toks) =
+parseType :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseType hkpos pos (CHSTokIdent _ ide:toks) =
   do
     toks' <- parseEndHook toks
     frags <- parseFrags toks'
-    return $ CHSHook (CHSType ide pos) : frags
-parseType _ toks = syntaxError toks
+    return $ CHSHook (CHSType ide pos) hkpos : frags
+parseType _ _ toks = syntaxError toks
 
-parseSizeof :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseSizeof pos (CHSTokIdent _ ide:toks) =
+parseSizeof :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseSizeof hkpos pos (CHSTokIdent _ ide:toks) =
   do
     toks' <- parseEndHook toks
     frags <- parseFrags toks'
-    return $ CHSHook (CHSSizeof ide pos) : frags
-parseSizeof _ toks = syntaxError toks
+    return $ CHSHook (CHSSizeof ide pos) hkpos : frags
+parseSizeof _ _ toks = syntaxError toks
 
-parseAlignof :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseAlignof pos (CHSTokIdent _ ide:toks) =
+parseAlignof :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseAlignof hkpos pos (CHSTokIdent _ ide:toks) =
   do
     toks' <- parseEndHook toks
     frags <- parseFrags toks'
-    return $ CHSHook (CHSAlignof ide pos) : frags
-parseAlignof _ toks = syntaxError toks
+    return $ CHSHook (CHSAlignof ide pos) hkpos : frags
+parseAlignof _ _ toks = syntaxError toks
 
-parseEnum :: Position -> [CHSToken] -> CST s [CHSFrag]
+parseEnum :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
 
 -- {#enum define hsid {alias_1,...,alias_n}  [deriving (clid_1,...,clid_n)] #}
-parseEnum pos (CHSTokIdent _ def: CHSTokIdent _ hsid: toks)
+parseEnum hkpos pos (CHSTokIdent _ def: CHSTokIdent _ hsid: toks)
   | identToString def == "define" =
   do
     (trans , toks')   <- parseTrans          toks
     (derive, toks'')  <- parseDerive         toks'
     toks'''           <- parseEndHook        toks''
     frags             <- parseFrags          toks'''
-    return $ CHSHook (CHSEnumDefine hsid trans derive pos) : frags
+    return $ CHSHook (CHSEnumDefine hsid trans derive pos) hkpos : frags
 
 -- {#enum cid [as hsid] {alias_1,...,alias_n}  [with prefix = pref] [deriving (clid_1,...,clid_n)] #}
-parseEnum pos (CHSTokIdent _ ide:toks) =
+parseEnum hkpos pos (CHSTokIdent _ ide:toks) =
   do
-    (oalias, toks' )   <- parseOptAs ide True toks
-    (trans , toks'')   <- parseTrans          toks'
-    (oprefix, toks''') <- parseOptPrefix True toks''
-    (derive, toks'''') <- parseDerive         toks'''
-    toks'''''          <- parseEndHook        toks''''
-    frags              <- parseFrags          toks'''''
-    return $ CHSHook (CHSEnum ide (norm oalias) trans oprefix derive pos) : frags
+    (oalias,      toks2) <- parseOptAs ide True toks
+    (trans,       toks3) <- parseTrans          toks2
+    (oprefix,     toks4) <- parseOptPrefix True toks3
+    (oreplprefix, toks5) <- parseOptReplPrefix  toks4
+    (derive,      toks6) <- parseDerive         toks5
+    toks7                <- parseEndHook        toks6
+    frags                <- parseFrags          toks7
+    return $ CHSHook (CHSEnum ide (norm oalias) trans
+                      oprefix oreplprefix derive pos) hkpos : frags
   where
     norm Nothing                   = Nothing
     norm (Just ide') | ide == ide' = Nothing
                      | otherwise   = Just ide'
-parseEnum _ toks = syntaxError toks
+parseEnum _ _ toks = syntaxError toks
 
-parseCall          :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseCall pos toks  =
+parseCall          :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseCall hkpos pos toks  =
   do
     (isPure  , toks'   ) <- parseIsPure          toks
     (isUnsafe, toks''  ) <- parseIsUnsafe        toks'
@@ -922,10 +976,10 @@ parseCall pos toks  =
     toks'''''            <- parseEndHook         toks''''
     frags                <- parseFrags           toks'''''
     return $
-      CHSHook (CHSCall isPure isUnsafe apath oalias pos) : frags
+      CHSHook (CHSCall isPure isUnsafe apath oalias pos) hkpos : frags
 
-parseFun          :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseFun pos toks  =
+parseFun          :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseFun hkpos pos toks  =
   do
     (isPure  , toks' ) <- parseIsPure          toks
     (isUnsafe, toks'2) <- parseIsUnsafe        toks'
@@ -938,7 +992,7 @@ parseFun pos toks  =
     frags              <- parseFrags           toks'8
     return $
       CHSHook
-        (CHSFun isPure isUnsafe apath oalias octxt parms parm pos) :
+        (CHSFun isPure isUnsafe apath oalias octxt parms parm pos) hkpos :
       frags
   where
     parseOptContext (CHSTokHSVerb _ ctxt:CHSTokDArrow _:toks') =
@@ -973,7 +1027,7 @@ parseIsUnsafe (CHSTokUnsafe _:toks) = return (True , toks)
 parseIsUnsafe toks                  = return (False, toks)
 
 apathToIdent :: CHSAPath -> Ident
-apathToIdent (CHSRoot ide) =
+apathToIdent (CHSRoot _ ide) =
     let lowerFirst (c:cs) = toLower c : cs
     in internalIdentAt (posOf ide) (lowerFirst $ identToString ide)
 apathToIdent (CHSDeref apath _) =
@@ -1020,16 +1074,24 @@ parseParm toks =
     parseOptMarshType toks' =
       return (CHSValArg, toks')
 
-parseField :: Position -> CHSAccess -> [CHSToken] -> CST s [CHSFrag]
-parseField pos access toks =
+parseField :: Position -> Position -> CHSAccess -> [CHSToken] -> CST s [CHSFrag]
+parseField hkpos pos access toks =
   do
     (path, toks') <- parsePath  toks
     toks''        <- parseEndHook toks'
     frags         <- parseFrags toks''
-    return $ CHSHook (CHSField access path pos) : frags
+    return $ CHSHook (CHSField access path pos) hkpos : frags
 
-parsePointer :: Position -> [CHSToken] -> CST s [CHSFrag]
-parsePointer pos toks =
+parseOffsetof :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseOffsetof hkpos pos toks =
+  do
+    (path, toks') <- parsePath toks
+    toks''        <- parseEndHook toks'
+    frags         <- parseFrags toks''
+    return $ CHSHook (CHSOffsetof path pos) hkpos : frags
+
+parsePointer :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parsePointer hkpos pos toks =
   do
     (isStar, ide, toks')          <-
       case toks of
@@ -1054,7 +1116,7 @@ parsePointer pos toks =
     return $
       CHSHook
        (CHSPointer
-         isStar ide (norm ide oalias) ptrType isNewtype oRefType emit pos)
+         isStar ide (norm ide oalias) ptrType isNewtype oRefType emit pos) hkpos
        : frags
   where
     parsePtrType :: [CHSToken] -> CST s (CHSPtrType, [CHSToken])
@@ -1066,8 +1128,8 @@ parsePointer pos toks =
     norm ide (Just ide') | ide == ide' = Nothing
                          | otherwise   = Just ide'
 
-parseClass :: Position -> [CHSToken] -> CST s [CHSFrag]
-parseClass pos (CHSTokIdent  _ sclassIde:
+parseClass :: Position -> Position -> [CHSToken] -> CST s [CHSFrag]
+parseClass hkpos pos (CHSTokIdent  _ sclassIde:
                 CHSTokDArrow _          :
                 CHSTokIdent  _ classIde :
                 CHSTokIdent  _ typeIde  :
@@ -1075,15 +1137,16 @@ parseClass pos (CHSTokIdent  _ sclassIde:
   do
     toks' <- parseEndHook toks
     frags <- parseFrags toks'
-    return $ CHSHook (CHSClass (Just sclassIde) classIde typeIde pos) : frags
-parseClass pos (CHSTokIdent _ classIde :
-                CHSTokIdent _ typeIde  :
-                toks)                     =
+    return $ CHSHook (CHSClass (Just sclassIde)
+                      classIde typeIde pos) hkpos : frags
+parseClass hkpos pos (CHSTokIdent _ classIde :
+                      CHSTokIdent _ typeIde  :
+                      toks)                     =
   do
     toks' <- parseEndHook toks
     frags <- parseFrags toks'
-    return $ CHSHook (CHSClass Nothing classIde typeIde pos) : frags
-parseClass _ toks = syntaxError toks
+    return $ CHSHook (CHSClass Nothing classIde typeIde pos) hkpos : frags
+parseClass _ _ toks = syntaxError toks
 
 parseOptLib :: [CHSToken] -> CST s (Maybe String, [CHSToken])
 parseOptLib (CHSTokLib    _    :
@@ -1107,6 +1170,16 @@ parseOptPrefix _     (CHSTokWith   _:toks) = syntaxError toks
 parseOptPrefix _     (CHSTokPrefix _:toks) = syntaxError toks
 parseOptPrefix _     toks                  = return (Nothing, toks)
 
+parseOptReplPrefix :: [CHSToken] -> CST s (Maybe String, [CHSToken])
+parseOptReplPrefix (CHSTokAdd   _    :
+                    CHSTokPrefix _    :
+                    CHSTokEqual  _    :
+                    CHSTokString _ str:
+                    toks)                = return (Just str, toks)
+parseOptReplPrefix (CHSTokAdd    _:toks) = syntaxError toks
+parseOptReplPrefix (CHSTokPrefix _:toks) = syntaxError toks
+parseOptReplPrefix toks                  = return (Nothing, toks)
+
 -- first argument is the identifier that is to be used when `^' is given and
 -- the second indicates whether the first character has to be upper case
 --
@@ -1155,12 +1228,19 @@ parsePath (CHSTokStar pos:toks) =
   do
     (path, toks') <- parsePath toks
     return (CHSDeref path pos, toks')
+parsePath (CHSTokStruct _pos:tok:toks) =
+  case keywordToIdent tok of
+    (CHSTokIdent _ ide) ->
+      do
+        (pathWithHole, toks') <- parsePath' toks
+        return (pathWithHole (CHSRoot True ide), toks')
+    _ -> syntaxError (tok:toks)
 parsePath (tok:toks) =
   case keywordToIdent tok of
     (CHSTokIdent _ ide) ->
       do
         (pathWithHole, toks') <- parsePath' toks
-        return (pathWithHole (CHSRoot ide), toks')
+        return (pathWithHole (CHSRoot False ide), toks')
     _ -> syntaxError (tok:toks)
 parsePath toks = syntaxError toks
 
