@@ -87,11 +87,11 @@
 --    transfers control to the following binding-hook lexer:
 --
 --      ident       -> letter (letter | digit | `\'')*
---      reservedid  -> `as' | `call' | `class' | `context' | `deriving'
+--      reservedid  -> `add' | `as' | `call' | `class' | `context' | `deriving'
 --                   | `enum' | `foreign' | `fun' | `get' | `lib'
 --                   | `downcaseFirstLetter'
 --                   | `newtype' | `nocode' | `pointer' | `prefix' | `pure'
---                   | `set' | `sizeof' | `stable' | `type'
+--                   | `set' | `sizeof' | `stable' | `struct' | `type'
 --                   | `underscoreToCase' | `upcaseFirstLetter' | `unsafe' |
 --                   | `with'
 --      reservedsym -> `{#' | `#}' | `{' | `}' | `,' | `.' | `->' | `='
@@ -204,7 +204,9 @@ data CHSToken = CHSTokArrow   Position          -- `->'
               | CHSTokRBrace  Position          -- `}'
               | CHSTokLParen  Position          -- `('
               | CHSTokRParen  Position          -- `)'
+              | CHSTokHook    Position          -- `{#'
               | CHSTokEndHook Position          -- `#}'
+              | CHSTokAdd     Position          -- `add'
               | CHSTokAs      Position          -- `as'
               | CHSTokCall    Position          -- `call'
               | CHSTokClass   Position          -- `class'
@@ -219,6 +221,7 @@ data CHSToken = CHSTokArrow   Position          -- `->'
               | CHSTokLib     Position          -- `lib'
               | CHSTokNewtype Position          -- `newtype'
               | CHSTokNocode  Position          -- `nocode'
+              | CHSTokOffsetof Position         -- `offsetof'
               | CHSTokPointer Position          -- `pointer'
               | CHSTokPrefix  Position          -- `prefix'
               | CHSTokPure    Position          -- `pure'
@@ -227,6 +230,7 @@ data CHSToken = CHSTokArrow   Position          -- `->'
               | CHSTokSizeof  Position          -- `sizeof'
               | CHSTokAlignof Position          -- `alignof'
               | CHSTokStable  Position          -- `stable'
+              | CHSTokStruct  Position          -- `struct'
               | CHSTokType    Position          -- `type'
               | CHSTok_2Case  Position          -- `underscoreToCase'
               | CHSTokUnsafe  Position          -- `unsafe'
@@ -237,7 +241,7 @@ data CHSToken = CHSTokArrow   Position          -- `->'
               | CHSTokHSQuot  Position String   -- quoted Haskell ('...')
               | CHSTokIdent   Position Ident    -- identifier
               | CHSTokHaskell Position String   -- verbatim Haskell code
-              | CHSTokCPP     Position String   -- pre-processor directive
+              | CHSTokCPP     Position String Bool -- pre-processor directive
               | CHSTokLine    Position          -- line pragma
               | CHSTokC       Position String   -- verbatim C code
               | CHSTokCtrl    Position Char     -- control code
@@ -257,6 +261,7 @@ instance Pos CHSToken where
   posOf (CHSTokLParen  pos  ) = pos
   posOf (CHSTokRParen  pos  ) = pos
   posOf (CHSTokEndHook pos  ) = pos
+  posOf (CHSTokAdd     pos  ) = pos
   posOf (CHSTokAs      pos  ) = pos
   posOf (CHSTokCall    pos  ) = pos
   posOf (CHSTokClass   pos  ) = pos
@@ -271,6 +276,7 @@ instance Pos CHSToken where
   posOf (CHSTokLib     pos  ) = pos
   posOf (CHSTokNewtype pos  ) = pos
   posOf (CHSTokNocode  pos  ) = pos
+  posOf (CHSTokOffsetof pos ) = pos
   posOf (CHSTokPointer pos  ) = pos
   posOf (CHSTokPrefix  pos  ) = pos
   posOf (CHSTokPure    pos  ) = pos
@@ -279,6 +285,7 @@ instance Pos CHSToken where
   posOf (CHSTokSizeof  pos  ) = pos
   posOf (CHSTokAlignof pos  ) = pos
   posOf (CHSTokStable  pos  ) = pos
+  posOf (CHSTokStruct  pos  ) = pos
   posOf (CHSTokType    pos  ) = pos
   posOf (CHSTok_2Case  pos  ) = pos
   posOf (CHSTokUnsafe  pos  ) = pos
@@ -289,7 +296,7 @@ instance Pos CHSToken where
   posOf (CHSTokHSQuot  pos _) = pos
   posOf (CHSTokIdent   pos _) = pos
   posOf (CHSTokHaskell pos _) = pos
-  posOf (CHSTokCPP     pos _) = pos
+  posOf (CHSTokCPP     pos _ _) = pos
   posOf (CHSTokLine    pos  ) = pos
   posOf (CHSTokC       pos _) = pos
   posOf (CHSTokCtrl    pos _) = pos
@@ -309,6 +316,7 @@ instance Eq CHSToken where
   (CHSTokLParen   _  ) == (CHSTokLParen   _  ) = True
   (CHSTokRParen   _  ) == (CHSTokRParen   _  ) = True
   (CHSTokEndHook  _  ) == (CHSTokEndHook  _  ) = True
+  (CHSTokAdd      _  ) == (CHSTokAdd      _  ) = True
   (CHSTokAs       _  ) == (CHSTokAs       _  ) = True
   (CHSTokCall     _  ) == (CHSTokCall     _  ) = True
   (CHSTokClass    _  ) == (CHSTokClass    _  ) = True
@@ -323,6 +331,7 @@ instance Eq CHSToken where
   (CHSTokLib      _  ) == (CHSTokLib      _  ) = True
   (CHSTokNewtype  _  ) == (CHSTokNewtype  _  ) = True
   (CHSTokNocode   _  ) == (CHSTokNocode   _  ) = True
+  (CHSTokOffsetof _  ) == (CHSTokOffsetof _  ) = True
   (CHSTokPointer  _  ) == (CHSTokPointer  _  ) = True
   (CHSTokPrefix   _  ) == (CHSTokPrefix   _  ) = True
   (CHSTokPure     _  ) == (CHSTokPure     _  ) = True
@@ -331,6 +340,7 @@ instance Eq CHSToken where
   (CHSTokSizeof   _  ) == (CHSTokSizeof   _  ) = True
   (CHSTokAlignof  _  ) == (CHSTokAlignof  _  ) = True
   (CHSTokStable   _  ) == (CHSTokStable   _  ) = True
+  (CHSTokStruct   _  ) == (CHSTokStruct   _  ) = True
   (CHSTokType     _  ) == (CHSTokType     _  ) = True
   (CHSTok_2Case   _  ) == (CHSTok_2Case   _  ) = True
   (CHSTokUnsafe   _  ) == (CHSTokUnsafe   _  ) = True
@@ -341,7 +351,7 @@ instance Eq CHSToken where
   (CHSTokHSQuot   _ _) == (CHSTokHSQuot   _ _) = True
   (CHSTokIdent    _ _) == (CHSTokIdent    _ _) = True
   (CHSTokHaskell  _ _) == (CHSTokHaskell  _ _) = True
-  (CHSTokCPP      _ _) == (CHSTokCPP      _ _) = True
+  (CHSTokCPP    _ _ _) == (CHSTokCPP    _ _ _) = True
   (CHSTokLine     _  ) == (CHSTokLine     _  ) = True
   (CHSTokC        _ _) == (CHSTokC        _ _) = True
   (CHSTokCtrl     _ _) == (CHSTokCtrl     _ _) = True
@@ -362,6 +372,7 @@ instance Show CHSToken where
   showsPrec _ (CHSTokLParen  _  ) = showString "("
   showsPrec _ (CHSTokRParen  _  ) = showString ")"
   showsPrec _ (CHSTokEndHook _  ) = showString "#}"
+  showsPrec _ (CHSTokAdd     _  ) = showString "add"
   showsPrec _ (CHSTokAs      _  ) = showString "as"
   showsPrec _ (CHSTokCall    _  ) = showString "call"
   showsPrec _ (CHSTokClass   _  ) = showString "class"
@@ -376,6 +387,7 @@ instance Show CHSToken where
   showsPrec _ (CHSTokLib     _  ) = showString "lib"
   showsPrec _ (CHSTokNewtype _  ) = showString "newtype"
   showsPrec _ (CHSTokNocode  _  ) = showString "nocode"
+  showsPrec _ (CHSTokOffsetof _ ) = showString "offsetof"
   showsPrec _ (CHSTokPointer _  ) = showString "pointer"
   showsPrec _ (CHSTokPrefix  _  ) = showString "prefix"
   showsPrec _ (CHSTokPure    _  ) = showString "pure"
@@ -384,6 +396,7 @@ instance Show CHSToken where
   showsPrec _ (CHSTokSizeof  _  ) = showString "sizeof"
   showsPrec _ (CHSTokAlignof _  ) = showString "alignof"
   showsPrec _ (CHSTokStable  _  ) = showString "stable"
+  showsPrec _ (CHSTokStruct  _  ) = showString "struct"
   showsPrec _ (CHSTokType    _  ) = showString "type"
   showsPrec _ (CHSTok_2Case  _  ) = showString "underscoreToCase"
   showsPrec _ (CHSTokUnsafe  _  ) = showString "unsafe"
@@ -394,7 +407,7 @@ instance Show CHSToken where
   showsPrec _ (CHSTokHSQuot  _ s) = showString ("'" ++ s ++ "'")
   showsPrec _ (CHSTokIdent   _ i) = (showString . identToString) i
   showsPrec _ (CHSTokHaskell _ s) = showString s
-  showsPrec _ (CHSTokCPP     _ s) = showString s
+  showsPrec _ (CHSTokCPP  _ s nl) = showString (if nl then "\n" else "" ++ s)
   showsPrec _ (CHSTokLine    _  ) = id            --TODO show line num?
   showsPrec _ (CHSTokC       _ s) = showString s
   showsPrec _ (CHSTokCtrl    _ c) = showChar c
@@ -465,6 +478,7 @@ chslexer  =      haskell        -- Haskell code
             >||< ctrl           -- control code (that has to be preserved)
             >||< hook           -- start of a binding hook
             >||< cpp            -- a pre-processor directive (or `#c')
+            >||< startmarker    -- marks beginning of input
 
 -- | stream of Haskell code (terminated by a control character or binding hook)
 --
@@ -586,7 +600,16 @@ ctrl  =
 --
 hook :: CHSLexer
 hook  = string "{#"
-        `lexmeta` \_ pos s -> (Nothing, incPos pos 2, s, Just bhLexer)
+        `lexmeta` \_ pos s -> (Just $ Right (CHSTokHook pos),
+                               incPos pos 2, s, Just bhLexer)
+
+-- | start marker: used to identify pre-processor directive at
+-- beginning of input -- this lexer just drops the start marker if it
+-- hasn't been used to handle a pre-processor directive
+--
+startmarker :: CHSLexer
+startmarker = char '\000' `lexmeta`
+              \lexeme pos s -> (Nothing, incPos pos 1, s, Just chslexer)
 
 -- | pre-processor directives and `#c'
 --
@@ -598,21 +621,23 @@ cpp :: CHSLexer
 cpp = directive
       where
         directive =
-          string "\n#" +> alt ('\t':inlineSet)`star` epsilon
+          (string "\n#" >|< string "\0#") +>
+          alt ('\t':inlineSet)`star` epsilon
           `lexmeta`
-             \(_:_:dir) pos s ->        -- strip off the "\n#"
+             \t@(ld:_:dir) pos s ->      -- strip off the "\n#" or "\0#"
                case dir of
                  ['c']                      ->          -- #c
-                   (Nothing, retPos pos, s, Just cLexer)
+                   (Nothing, incPos pos (length t), s, Just cLexer)
                  -- a #c may be followed by whitespace
                  'c':sp:_ | sp `elem` " \t" ->          -- #c
-                   (Nothing, retPos pos, s, Just cLexer)
+                   (Nothing, incPos pos (length t), s, Just cLexer)
                  ' ':line@(n:_) | isDigit n ->                 -- C line pragma
                    let pos' = adjustPosByCLinePragma line pos
                     in (Just $ Right (CHSTokLine pos'), pos', s, Nothing)
                  _                            ->        -- CPP directive
-                   (Just $ Right (CHSTokCPP pos dir),
-                    retPos pos, s, Nothing)
+                   (Just $ Right (CHSTokCPP pos dir (ld == '\n')),
+                    if ld == '\n' then retPos pos else incPos pos (length t),
+                    s, Nothing)
 
 adjustPosByCLinePragma :: String -> Position -> Position
 adjustPosByCLinePragma str pos = adjustPos fname' row' pos
@@ -684,6 +709,7 @@ identOrKW  =
        (letter +> (letter >|< digit >|< char '\'')`star` epsilon
        `lexactionName` \cs pos name -> (idkwtok $!pos) cs name)
   where
+    idkwtok pos "add"              _    = CHSTokAdd     pos
     idkwtok pos "as"               _    = CHSTokAs      pos
     idkwtok pos "call"             _    = CHSTokCall    pos
     idkwtok pos "class"            _    = CHSTokClass   pos
@@ -698,6 +724,7 @@ identOrKW  =
     idkwtok pos "lib"              _    = CHSTokLib     pos
     idkwtok pos "newtype"          _    = CHSTokNewtype pos
     idkwtok pos "nocode"           _    = CHSTokNocode  pos
+    idkwtok pos "offsetof"         _    = CHSTokOffsetof pos
     idkwtok pos "pointer"          _    = CHSTokPointer pos
     idkwtok pos "prefix"           _    = CHSTokPrefix  pos
     idkwtok pos "pure"             _    = CHSTokPure    pos
@@ -706,6 +733,7 @@ identOrKW  =
     idkwtok pos "sizeof"           _    = CHSTokSizeof  pos
     idkwtok pos "alignof"          _    = CHSTokAlignof pos
     idkwtok pos "stable"           _    = CHSTokStable  pos
+    idkwtok pos "struct"           _    = CHSTokStruct  pos
     idkwtok pos "type"             _    = CHSTokType    pos
     idkwtok pos "underscoreToCase" _    = CHSTok_2Case  pos
     idkwtok pos "unsafe"           _    = CHSTokUnsafe  pos
@@ -718,6 +746,7 @@ identOrKW  =
 keywordToIdent :: CHSToken -> CHSToken
 keywordToIdent tok =
   case tok of
+    CHSTokAdd     pos -> mkid pos "add"
     CHSTokAs      pos -> mkid pos "as"
     CHSTokCall    pos -> mkid pos "call"
     CHSTokClass   pos -> mkid pos "class"
@@ -732,6 +761,7 @@ keywordToIdent tok =
     CHSTokLib     pos -> mkid pos "lib"
     CHSTokNewtype pos -> mkid pos "newtype"
     CHSTokNocode  pos -> mkid pos "nocode"
+    CHSTokOffsetof pos -> mkid pos "offsetof"
     CHSTokPointer pos -> mkid pos "pointer"
     CHSTokPrefix  pos -> mkid pos "prefix"
     CHSTokPure    pos -> mkid pos "pure"
@@ -740,6 +770,7 @@ keywordToIdent tok =
     CHSTokSizeof  pos -> mkid pos "sizeof"
     CHSTokAlignof pos -> mkid pos "alignof"
     CHSTokStable  pos -> mkid pos "stable"
+    CHSTokStruct  pos -> mkid pos "struct"
     CHSTokType    pos -> mkid pos "type"
     CHSTok_2Case  pos -> mkid pos "underscoreToCase"
     CHSTokUnsafe  pos -> mkid pos "unsafe"
@@ -797,7 +828,7 @@ inhsverb  = alt ([' '..'\127'] \\ "'")
 -- | character sets
 --
 anySet, inlineSet, specialSet, commentSpecialSet, ctrlSet :: [Char]
-anySet            = ['\0'..'\255']
+anySet            = ['\1'..'\255']
 inlineSet         = anySet \\ ctrlSet
 specialSet        = ['{', '-', '"', '\'']
 commentSpecialSet = ['{', '-']
@@ -819,7 +850,7 @@ lexCHS cs pos  =
   do
     nameSupply <- getNameSupply
     state <- initialState nameSupply
-    let (ts, lstate, errs) = execLexer chslexer (cs, pos, state)
+    let (ts, lstate, errs) = execLexer chslexer ('\0':cs, pos, state)
         (_, pos', state')  = lstate
     mapM_ raise errs
     assertFinalState pos' state'
