@@ -168,7 +168,7 @@ ghFrag (frag@(CHSVerb  _ _  ) : frags) =
 
 -- generate an enum __c2hs__enum__'id { __c2hs_enr__'id = DEF1, ... } and then process an
 -- ordinary enum directive
-ghFrag (_frag@(CHSHook (CHSEnumDefine hsident trans instances pos)) : frags) =
+ghFrag (_frag@(CHSHook (CHSEnumDefine hsident trans instances pos) hkpos) : frags) =
   do ide <- newEnumIdent
      (enrs,trans') <- createEnumerators trans
      return (DL.open [show.pretty $ enumDef ide enrs,";\n"], Frag (enumFrag (identToString ide) trans'), frags)
@@ -184,9 +184,9 @@ ghFrag (_frag@(CHSHook (CHSEnumDefine hsident trans instances pos)) : frags) =
   createEnumerator (cid,hsid) = liftM (\enr -> ((enr,cid),(enr,hsid))) newEnrIdent
   enumDef ide enrs = CEnum (Just ide) (Just$ map mkEnr enrs) [] undefNode
     where mkEnr (name,value) = (name, Just $ CVar value undefNode)
-  enumFrag ide trans' = CHSHook (CHSEnum (internalIdent ide) (Just hsident) trans' Nothing instances pos)
+  enumFrag ide trans' = CHSHook (CHSEnum (internalIdent ide) (Just hsident) trans' Nothing Nothing instances pos) hkpos
 
-ghFrag (frag@(CHSHook  _    ) : frags) =
+ghFrag (frag@(CHSHook  _    _) : frags) =
   return (DL.zero, Frag frag, frags)
 ghFrag (frag@(CHSLine  _    ) : frags) =
   return (DL.zero, Frag frag, frags)
@@ -198,7 +198,7 @@ ghFrag (     (CHSC    s  _  ) : frags) =
 
 ghFrag (     (CHSCond _  _  ) : _    ) =
   interr "GenHeader.ghFrags: There can't be a structured conditional yet!"
-ghFrag (     (CHSCPP  s  pos) : frags) =
+ghFrag (     (CHSCPP s pos nl) : frags) =
   let
     (directive, _) =   break (`elem` " \t")
                      . dropWhile (`elem` " \t")
@@ -211,7 +211,8 @@ ghFrag (     (CHSCPP  s  pos) : frags) =
     "else"   -> return (DL.zero              , Else   pos               , frags)
     "elif"   -> return (DL.zero              , Elif s pos               , frags)
     "endif"  -> return (DL.zero              , Endif  pos               , frags)
-    _        -> return (DL.open ['#':s, "\n"], Frag   (CHSVerb "" nopos), frags)
+    _        -> return (DL.open ['#':s, "\n"],
+                        Frag (CHSVerb (if nl then "\n" else "") pos), frags)
   where
     -- enter a new conditional (may be an #if[[n]def] or #elif)
     --
@@ -250,7 +251,7 @@ ghFrag (     (CHSCPP  s  pos) : frags) =
                              _                       ->
                                interr "GenHeader.ghFrag: Expected CHSCond!"
           Endif   _   -> closeIf (headerTh `DL.snoc` "#endif\n")
-                                 (s, fragsTh)
+                                 (s', fragsTh)
                                  []
                                  (Just [])
                                  rest
