@@ -69,8 +69,8 @@
 --
 --  -i DIRS
 --  --include=DIRS
---        Search the colon separated list of directories DIRS when searching
---        for .chi files.
+--        Search the colon (Linux) or semicolon (Windows) separated
+--        list of directories DIRS when searching for .chi files.
 --
 --  -k
 --  --keep
@@ -125,7 +125,7 @@ import System.Console.GetOpt
                   (ArgOrder(..), OptDescr(..), ArgDescr(..), usageInfo, getOpt)
 import qualified System.FilePath as FilePath
                   (takeExtension, dropExtension, takeBaseName)
-import System.FilePath ((<.>), (</>))
+import System.FilePath ((<.>), (</>), splitSearchPath)
 import System.IO (stderr, openFile, IOMode(..))
 import System.IO.Error (ioeGetErrorString, ioeGetFileName)
 import System.Process (runProcess, waitForProcess)
@@ -178,6 +178,8 @@ errTrailer = "Try the option `--help' on its own for more information.\n"
 --
 data Flag = CPPOpts  String     -- ^ additional options for C preprocessor
           | CPP      String     -- ^ program name of C preprocessor
+          | NoGNU               -- ^ suppress GNU preprocessor symbols
+          | NoBlocks            -- ^ suppress MacOS __BLOCKS__ preproc. symbol
           | Dump     DumpType   -- ^ dump internal information
           | Help                -- ^ print brief usage information
           | Keep                -- ^ keep the .i file
@@ -209,6 +211,14 @@ options  = [
          ["cpp"]
          (ReqArg CPP "CPP")
          "use executable CPP to invoke C preprocessor",
+  Option ['n']
+         ["no-gnu"]
+         (NoArg NoGNU)
+         "suppress GNU preprocessor symbols",
+  Option ['b']
+         ["no-blocks"]
+         (NoArg NoBlocks)
+         "suppress MacOS __BLOCKS__ preprocessor symbol",
   Option ['d']
          ["dump"]
          (ReqArg dumpArg "TYPE")
@@ -393,6 +403,8 @@ help =
 processOpt :: Flag -> CST s ()
 processOpt (CPPOpts  cppopt ) = addCPPOpts  [cppopt]
 processOpt (CPP      cpp    ) = setCPP      cpp
+processOpt (NoGNU           ) = setNoGNU
+processOpt (NoBlocks        ) = setNoBlocks
 processOpt (Dump     dt     ) = setDump     dt
 processOpt (Keep            ) = setKeep
 processOpt (Library         ) = setLibrary
@@ -453,6 +465,16 @@ addCPPOpts opts  = setSwitch $ \sb -> sb {cppOptsSB = cppOptsSB sb ++ opts}
 setCPP       :: FilePath -> CST s ()
 setCPP fname  = setSwitch $ \sb -> sb {cppSB = fname}
 
+-- | set flag to suppress GNU preprocessor symbols
+--
+setNoGNU :: CST s ()
+setNoGNU  = setSwitch $ \sb -> sb {noGnuSB = True}
+
+-- | set flag to suppress MacOS __BLOCKS__ preprocessor symbols
+--
+setNoBlocks :: CST s ()
+setNoBlocks = setSwitch $ \sb -> sb {noBlocksSB = True}
+
 -- set the given dump option
 --
 setDump         :: DumpType -> CST s ()
@@ -479,18 +501,8 @@ setLibrary  = setSwitch $ \sb -> sb {librarySB = True}
 --   standard value in the compiler state.
 --
 setInclude :: String -> CST s ()
-setInclude str = do
-  let fp = makePath str ""
-  setSwitch $ \sb -> sb {chiPathSB = fp ++ (chiPathSB sb)}
-  where
-    makePath ('\\':r:em)   path = makePath em (path ++ ['\\',r])
-    makePath (' ':r)       path = makePath r path
-    makePath (':':r)       ""   = makePath r ""
-    makePath (':':r)       path = path : makePath r ""
-    makePath ('/':':':r)   path = path : makePath r ""
-    makePath (r:emain)     path = makePath emain (path ++ [r])
-    makePath ""            ""   = []
-    makePath ""            path = [path]
+setInclude str =
+  setSwitch $ \sb -> sb {chiPathSB = splitSearchPath str ++ (chiPathSB sb)}
 
 -- | set the output file name
 --
@@ -572,9 +584,16 @@ process headerFiles bndFile  =
     --
     -- run C preprocessor over the header
     --
-    cpp     <- getSwitch cppSB
-    cppOpts <- getSwitch cppOptsSB
-    let args = cppOpts ++ [newHeaderFile]
+    cpp      <- getSwitch cppSB
+    cppOpts  <- getSwitch cppOptsSB
+    noGnu    <- getSwitch noGnuSB
+    noBlocks <- getSwitch noBlocksSB
+    let noGnuOpts =
+          if noGnu
+          then ["-U__GNUC__", "-U__GNUC_MINOR__", "-U__GNUC_PATCHLEVEL__"]
+          else []
+        noBlocksOpts = if noBlocks then ["-U__BLOCKS__"] else []
+        args = cppOpts ++ noGnuOpts ++ noBlocksOpts ++ [newHeaderFile]
     tracePreproc (unwords (cpp:args))
     exitCode <- CIO.liftIO $ do
       preprocHnd <- openFile preprocFile WriteMode
